/***************************************************************************
                          soundwrapper.cpp  -  description
                             -------------------
    begin                : Sam Mai 25 2002
    copyright            : (C) 2002, 2004 by Thomas Friedrichsmeier
    email                : tfry@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "soundwrapper.h"

#include <arts/kmedia2.h>
#include <arts/artsflow.h>
#include <arts/qiomanager.h>
#include <klocale.h>
#include <arts/kartsdispatcher.h>
#include <arts/kartsserver.h>

#include "cdp.h"

KArtsDispatcher dispatcher;

SoundWrapper::SoundWrapper(){
	sound_error = false;
	sound_enabled = false;
	sound = false;
}

SoundWrapper::~SoundWrapper(){
	if (sound) {
		stop_server ();
	}
}

void SoundWrapper::init(bool enable_sound) {

	sound_enabled = enable_sound;

	if (!sound_enabled) {
		if (sound) {
			stop_server ();
		}
	} else {
		if (!sound) {
			start_server ();
		}
	}

	sound = sound_enabled && (!sound_error);
}

void SoundWrapper::start_sound (int id, float pitch) {
	if (sound) {
		if (id) {
       		if (sound_map.contains (id)) {
				Sound this_sound = sound_map[id];
				if (this_sound.created) {
       				this_sound.player.stop ();		// if this isn't done in a while (i.e. stop_sound was not called), sound
												// keeps getting louder (strange thing, no?)
					this_sound.player.start ();
					if (this_sound.continuous) {
						this_sound.po.loop (true);
					}
					this_sound.po.speed (pitch);
					this_sound.po.start ();
					this_sound.po.play ();
				}
			}
		}
	}
}

void SoundWrapper::stop_sound (int id) {
	if (sound) {
		if (id) {
			if (sound_map.contains (id)) {
				Sound this_sound = sound_map[id];
				if (this_sound.created) {
					if (this_sound.immediate_interrupt) {
						this_sound.player.stop ();
						this_sound.po.halt ();
						this_sound.po.stop ();
					} else {
						this_sound.po.loop (false);
					}
				}
			}			
		}
	}
}

int SoundWrapper::register_sound (const QString &name, bool continuous, bool immediate_interrupt) {

	Sound new_sound;
	new_sound.filename = name;
	new_sound.created = false;
	new_sound.continuous = continuous;
	new_sound.immediate_interrupt = immediate_interrupt;

	int id;       			// determine id and assign
	if (freed_ids.empty ()) {
		id = sound_map.count () + 1;
	} else {                      // if we have unregistered sounds before, we need to make sure not to assign ids already in use, but rather use previously freed ids
		id = freed_ids.first ();
		freed_ids.pop_front ();
	}
	sound_map.insert (id, new_sound);

	if (sound) {			// create sound if sound is enabled
		create_sound (sound_map.find (id));
	}

	return (id);			// return id
}

void SoundWrapper::unregister_sound (int id) {
	if (id) {
		if (sound_map.contains (id)) {
			sound_map.remove (id);
			freed_ids.append (id);
			if (sound_map.empty ()) {
				freed_ids.clear ();
			}
		}
	} 	
}

/** Creates server, factory and all registered sounds */
void SoundWrapper::start_server () {
	KArtsServer kserver;
	server = kserver.server();
	if (server.isNull ()) {
		sound_error = true;
		cdp->debug_msg (i18n ("Could not reference sound-server. Sound is disabled."), Serious);
	} else {
		server.checkNewObjects ();

		// (re)create PlayObjects
		SoundMap::Iterator it;
		for ( it = sound_map.begin(); it != sound_map.end(); ++it ) {
			create_sound (it);
		}	
	}
}

/** Discards server, factory and all created registered sounds */
void SoundWrapper::stop_server () {
	// discard created sounds
	SoundMap::Iterator it;
	for ( it = sound_map.begin(); it != sound_map.end(); ++it ) {
		discard_sound (it);
	}	

	// discard server
	server = SoundServerV2::null ();
}

/** Creates a single sound */
void SoundWrapper::create_sound (SoundMap::Iterator it) {
	bool ok = false;
	it.data ().po = Arts::DynamicCast(server.createObject("EXT_WavPlayObject"));
	it.data ().player = Arts::DynamicCast(server.createObject("Arts::Synth_AMAN_PLAY"));
	if (!it.data ().po.isNull () && !it.data ().player.isNull ()) {
		if (it.data ().filename.latin1 () != bad_string_value) {  // if it is, we have alread complained about this in cdp->prefix_check
			ok = it.data ().po.loadMedia (it.data ().filename.latin1 ());
		}
		if (ok) {
			QString dummy;                        // create unique title and autoRestoreID
			dummy.setNum (it.key ());		
			dummy += "tp";
			dummy += it.data ().filename;
			it.data ().player.title(dummy.latin1 ());
			it.data ().player.autoRestoreID(dummy.latin1 ());

			Arts::connect (it.data ().po, "left", it.data ().player, "left");            // connect
			Arts::connect (it.data ().po, "right", it.data ().player, "right");
		}
	}
	if (ok) {
		it.data ().created = true;		
	} else {
		cdp->debug_msg (i18n ("Error reading sound file '%1'. Sound will not be played.").arg(it.data().filename), Warn);
	}

}

/** Discards a single (created, registered) sound */
void SoundWrapper::discard_sound (SoundMap::Iterator it) {
	if (it.data ().created) {
		it.data ().created = false;
		it.data ().po = EXT_WavPlayObject::null ();
		it.data ().player = Synth_AMAN_PLAY::null ();
	}
}
