/***************************************************************************
                          taxi.h  -  description
                             -------------------
    begin                : Fri May 18 2001
    copyright            : (C) 2001, 2002, 2004, 2005 by Thomas Friedrichsmeier
    email                : Thomas.Friedrichsmeier@ruhr-uni-bochum.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TAXI_H
#define TAXI_H

#include <qstring.h>

#include "timedobject.h"
#include "tpbase.h"
#include "soundhandle.h"
#include "animatedsprite.h"

/**
  *@author Thomas Friedrichsmeier
  */

class QDomElement;
class Teleporter;
class Platform;
class SoundWrapper;
class Animation;

/** This class is basically the taxi-sprite. In addition it handles sprites for its gear (optional) and for
	its flying_exhaust. The main communication-partner for this class is Taxipilot, but sometimes it directly
	communicates with teleporters and (moving) platforms. Usually, there is only ever one instance of this
	class existent at any point of time (referenced by Taxipilot::taxi), but that instance does not necessarily
	remain the same throughout runtime. */

class Taxi: public AnimatedSprite, public TimedObject {
  public:
	Taxi(const QString & file, int base_x = 0, int base_y = 0, bool for_credits_only = false);
	~Taxi();
/** Signals we are going down. See up () for details. */
	void down(float acc);
/** Signals we are going up, sets speed/animation/sound accordingly. The float acceleration is passed as
a parameter instead of using some constant amount, so we may eventually
abstract from keyboard input, and allow (analog) joystick input */
	void up(float acc);
/** See above */
	void right(float acc);
/** See above */
	void left(float acc);
/** Stops the currently running flying animation (and esp. the sound(s)), if any. (called from Taxipilot::do_chores ()) */
	void stop_flying_animation();
/** Takes care of teleportation (appearing, disappearing animations), setting new position
and speed. If appear_only is given, does not do disappearing animation. If the target is moving, you should
pass a non-zero pointer to the target teleporter. */
	void teleport_to(int x, int y, float speedfactor_x, float speedfactor_y, bool appear_only = false, Teleporter *dynamic = 0);
/** Returns the current fuel level in percent of the taxi (Used by Taxipilot to set the
FuelDial) */
	int fuel_state();
/** These two are set by fuel_state () and should be self-explanatory */
	bool fuel_low, fuel_out;
/** Increases the taxi's fuel supply by step (max up to max_fuel). Returns actual
increase in fuel (so that's 0 if taxi is already full). */
	float fuel_up(int step);
/** Toggles the taxis landing-gear */
	void toggle_gear();
/** Repositions the taxi. This also takes care of repositioning the landing gear. */
	void reposition(int x, int y, float xv, float yv, bool align_bottom = false);;
/** Returns the (absolute) x-position the left landing-gear is to be found at */
	int landing_gear_x1();
/** Returns the (absolute) x-position the right landing-flap is to be found at */
	int landing_gear_x2();
/** A wrapper for moveBy, taking care of the gear if appropriate */
	void wrap_moveBy (float delta_x, float delta_y);
/** Adjusts the taxi to a fractional position, so it doesn't wiggle on a moving platform */
	void adjust (float fraction_x, float fraction_y);
/** Returns the (absolute) y-position the landing-gear are to be found at */
	int landing_gear_y();
/** Sprite for the gear-animation */
	AnimatedSprite *gear;

/** These additional sprites are needed for the flight-exhausts (harmful to passenger, no crashing
into levels */
	AnimatedSprite *flying_lr, *flying_ud;
/** Is taxi ready for flight (not teleporting) */
	bool ready;
/** x-offsets of the landing gear, used to check, whether taxi is properly landed within platform boundaries */
	int landing_gear_off_x1[2], landing_gear_off_x2[2];

/** Initializes the taxi to have fuel fuel and a fuel-capacity max_fuel */
	void initialize (float fuel, float max_fuel);

	friend class Taxipilot;		// we need to know absX2 and stuff

/** about data */
	QString title, author;
/** Current status of the landing gear: 0=in 1=half in/out 2=out. */
	int gear_state;
/** The previous state of the gear (i.e. before crash / teleport). Can only be in (flase) or out (true) */
	bool gear_prev_state;
/** Are gear available for this taxi? */
	bool gear_avail;
/** Is the taxi currently landed? */
	bool landed;
private:						// Private methods
	float speed;				// store how fast we were going last.
/** Store the last speed when teleporting */
	float speed_x, speed_y;
/** Advance timer */
	int advance_timer;
/** Current fuel state */
	float fuel;
/** Maximum fuel state */
	float max_fuel;

/** This function reads a taxi-description ( specs for the animations ) from a file.
read_taxi is only called by the constructor. */
	void read_taxi(const QString & file);
/** Tell the taxi it has landed on the specified paltform */
	void land (Platform *on);
/** Pointer to the platform the taxi is landed on (if it is landed) */
	Platform *platform_p;
/** stores the timer-id for triggering and animating the idle-animations, respectively.
"Idle" animation is a slightly misleading term, since the taxi will do these animations
regardless of its own activity (at least for now) */
	int idle_timer;
/** The number of different idle_animations this taxi has. Must be at least one. */
	unsigned int num_idle_animations;

/** True, when the taxi is teleporting through a dynamic teleporter, i.e. one whose target is moving. */
	bool dynamic_teleport;
/** And in the above case, this stores a pointer to the target teleporter */
	Teleporter *dynamic_teleporter;

/** These store the animations the taxi (and it associated sprites) uses */
	Animation *gear_anim, *still, *teleporting, **idle, *flying_left, *flying_right, *flying_up, *flying_down;

/** When teleporting, these values have to be stored temporarily */
	int new_x, new_y;
	float new_speed_x, new_speed_y;
/** just so we don't have to calculate (width () / 2) and (height () / 2) all the time */
	int off_x, off_y;
	int twice_off_y;
/** Whether left/right-acceleration is blocked when gear are out */
	bool gear_block;

/** Sets of the animation anim for this taxi or it's gear, and triggers
a function-call when completed. Reverse means: play in reverse order; gear means, the animation is for the gear */
	void do_animation (Animation * anim, void (Taxi::*call_back) (), bool reverse=false, bool gear=false, bool no_sound=false);

/** Stores, which direction(s) we were last flying in. Mostly needed for starting and stopping sounds. */
	int flying_state, prev_flying_state;
	enum Flying_states { Up=1, Left=2, Down=4, Right=8 };
/** Flying-animation lag up/down-left/right */
	float flying_lag_x, flying_lag_y;
#define MAX_FLYING_LAG 6.0
/** further sounds that can't go with generalized animations (Anim-struct) */
	SoundHandle crash_sound, land_sound;

/** Timers for the animations (of the taxi itself and of its gear) */
	int gear_anim_timer, anim_timer;
/** Callbacks called when animations are done */
	void (Taxi::*taxi_call_back) ();
	void (Taxi::*gear_call_back) ();
/** Taxi is to reappear (at global new_x, new_y) */
	void reappear ();
/** Stuff to do when gear have just finished going in */
	void gear_in ();
/** Stuff to do when the gear just finished going out */
	void gear_out();
/** Starts idle_timer (not idle_animation) */
	void go_idle ();
/** Does nothing */
	void do_nothing();

/** Stuff to do when done appearing */
	void done_appearing ();

/** Tells the taxi it crashed. Does stuff like setting of an appropriate animation,
sound and other things to do. hard specifies, whether the taxi crashed into something hard (as opposed to
into the passenger) and whether the crash_sound will be played */
	void crash(bool hard);
/** The config has changed. Recalculate. */
	void config_changed();
/** Fuel-consumption (dependend on cdp->cps ()) */
	float fuel_con;
/** Flying-animation lagging step (dependend on cdp->cps ()) */
	float flying_lag_step;
protected:					// Protected methods
/** Mostly everyting will be done inside: timing for all animations */
	void timerTick(int id);
};

#endif
